@@ -26,87 +26,14 @@ import pan.commit
 import panos
 import panos.errors as err
 from panos import base, firewall, getlogger, policies, yesno
-from panos.base import ENTRY, MEMBER, PanObject, Root
+from panos.base import ENTRY, MEMBER, OpState, PanObject, Root
 from panos.base import VarPath as Var
 from panos.base import VersionedPanObject, VersionedParamPath
 
 logger = getlogger(__name__)
 
 
-class DeviceGroup(VersionedPanObject):
-    """Panorama Device-group
-
-    This class and the :class:`panos.panorama.Panorama` classes are the only objects that can
-    have a :class:`panos.firewall.Firewall` child object. In addition to a Firewall, a
-    DeviceGroup can have the same children objects as a :class:`panos.firewall.Firewall`
-    or :class:`panos.device.Vsys`.
-
-    See also :ref:`classtree`
-
-    Args:
-        name (str): Name of the device-group
-        tag (list): Tags as strings
-
-    """
-
-    ROOT = Root.DEVICE
-    SUFFIX = ENTRY
-    VSYS_LABEL = "device-group"
-    CHILDTYPES = (
-        "firewall.Firewall",
-        "objects.AddressObject",
-        "objects.AddressGroup",
-        "objects.ServiceObject",
-        "objects.ServiceGroup",
-        "objects.ApplicationObject",
-        "objects.ApplicationGroup",
-        "objects.ApplicationFilter",
-        "objects.ScheduleObject",
-        "objects.SecurityProfileGroup",
-        "objects.CustomUrlCategory",
-        "objects.LogForwardingProfile",
-        "objects.Region",
-        "objects.Edl",
-        "policies.PreRulebase",
-        "policies.PostRulebase",
-    )
-
-    def _setup(self):
-        # xpaths
-        self._xpaths.add_profile(value="/device-group")
-
-        # params
-        params = []
-
-        params.append(VersionedParamPath("tag", vartype="entry"))
-
-        self._params = tuple(params)
-
-    def _setup_opstate(self):
-        self.opstate = DeviceGroupOpState(self)
-
-    @property
-    def vsys(self):
-        return self.name
-
-    def devicegroup(self):
-        return self
-
-    def xpath_vsys(self):
-        return self.xpath()
-
-    def _setup_opstate(self):
-        self.opstate = DeviceGroupOpState(self)
-
-
-class DeviceGroupOpState(object):
-    """Operational state handling for device group classes."""
-
-    def __init__(self, obj):
-        self.dg_hierarchy = DeviceGroupHierarchy(obj)
-
-
-class DeviceGroupHierarchy(object):
+class DeviceGroupHierarchy(OpState):
     """Operational state handling for device group hierarchy.
 
     Args:
@@ -114,8 +41,7 @@ class DeviceGroupHierarchy(object):
 
     """
 
-    def __init__(self, obj):
-        self.obj = obj
+    def _setup(self):
         self.parent = None
 
     def refresh(self):
@@ -158,6 +84,70 @@ class DeviceGroupHierarchy(object):
         return dev.syncjob(resp)
 
 
+class DeviceGroup(VersionedPanObject):
+    """Panorama Device-group
+
+    This class and the :class:`panos.panorama.Panorama` classes are the only objects that can
+    have a :class:`panos.firewall.Firewall` child object. In addition to a Firewall, a
+    DeviceGroup can have the same children objects as a :class:`panos.firewall.Firewall`
+    or :class:`panos.device.Vsys`.
+
+    See also :ref:`classtree`
+
+    Args:
+        name (str): Name of the device-group
+        tag (list): Tags as strings
+
+    """
+
+    ROOT = Root.DEVICE
+    SUFFIX = ENTRY
+    VSYS_LABEL = "device-group"
+    CHILDTYPES = (
+        "firewall.Firewall",
+        "objects.AddressObject",
+        "objects.AddressGroup",
+        "objects.ServiceObject",
+        "objects.ServiceGroup",
+        "objects.ApplicationObject",
+        "objects.ApplicationGroup",
+        "objects.ApplicationTag",
+        "objects.ApplicationFilter",
+        "objects.ScheduleObject",
+        "objects.SecurityProfileGroup",
+        "objects.CustomUrlCategory",
+        "objects.LogForwardingProfile",
+        "objects.Region",
+        "objects.Edl",
+        "policies.PreRulebase",
+        "policies.PostRulebase",
+    )
+    OPSTATES = {
+        "dg_hierarchy": DeviceGroupHierarchy,
+    }
+
+    def _setup(self):
+        # xpaths
+        self._xpaths.add_profile(value="/device-group")
+
+        # params
+        params = []
+
+        params.append(VersionedParamPath("tag", vartype="entry"))
+
+        self._params = tuple(params)
+
+    @property
+    def vsys(self):
+        return self.name
+
+    def devicegroup(self):
+        return self
+
+    def xpath_vsys(self):
+        return self.xpath()
+
+
 class Template(VersionedPanObject):
     """A panorama template.
 
@@ -176,12 +166,15 @@ class Template(VersionedPanObject):
     SUFFIX = ENTRY
     CHILDTYPES = (
         "device.Vsys",
+        "device.VsysResources",
         "device.SystemSettings",
         "device.LogSettingsSystem",
         "device.LogSettingsConfig",
         "device.PasswordProfile",
         "device.Administrator",
         "device.SslDecrypt",
+        "device.LocalUserDatabaseUser",
+        "device.LocalUserDatabaseGroup",
         "ha.HighAvailability",
         "network.EthernetInterface",
         "network.AggregateInterface",
@@ -197,6 +190,11 @@ class Template(VersionedPanObject):
         "network.IpsecCryptoProfile",
         "network.IkeCryptoProfile",
         "network.GreTunnel",
+        "network.Zone",
+        "network.IpsecTunnelIpv4ProxyId",
+        "network.IpsecTunnelIpv6ProxyId",
+        "network.Layer2Subinterface",
+        "network.Layer3Subinterface",
         "panorama.TemplateVariable",
     )
 
@@ -280,6 +278,11 @@ class TemplateStack(VersionedPanObject):
         "network.IpsecCryptoProfile",
         "network.IkeCryptoProfile",
         "network.GreTunnel",
+        "network.Zone",
+        "network.IpsecTunnelIpv4ProxyId",
+        "network.IpsecTunnelIpv6ProxyId",
+        "network.Layer2Subinterface",
+        "network.Layer3Subinterface",
         "panorama.TemplateVariable",
     )
 
@@ -362,6 +365,33 @@ class TemplateVariable(VersionedPanObject):
         self._params = tuple(params)
 
 
+class PanoramaDeviceGroupHierarchy(OpState):
+    """Operational state handling for device group hierarchy."""
+
+    def fetch(self):
+        """Returns a dict of device groups and their parents.
+
+        Keys in the dict are the device group's name, while the value is the
+        name of that device group's parent.  Top level device groups will have
+        a parent of ``None``.
+
+        Returns:
+            dict
+
+        """
+
+        resp = self.obj.op("show dg-hierarchy")
+        data = resp.find("./result/dg-hierarchy")
+
+        ans = {}
+        nodes = [(None, x) for x in data.findall("./dg")]
+        for parent, elm in iter(nodes):
+            ans[elm.attrib["name"]] = parent
+            nodes.extend((elm.attrib["name"], x) for x in elm.findall("./dg"))
+
+        return ans
+
+
 class Panorama(base.PanDevice):
     """Panorama device
 
@@ -392,11 +422,33 @@ class Panorama(base.PanDevice):
         "device.HttpServerProfile",
         "device.CertificateProfile",
         "device.SslDecrypt",
+        "objects.AddressObject",
+        "objects.AddressGroup",
+        "objects.ServiceObject",
+        "objects.ServiceGroup",
+        "objects.Tag",
+        "objects.ApplicationObject",
+        "objects.ApplicationGroup",
+        "objects.ApplicationTag",
+        "objects.ApplicationFilter",
+        "objects.ApplicationContainer",
+        "objects.ScheduleObject",
+        "objects.SecurityProfileGroup",
+        "objects.CustomUrlCategory",
+        "objects.LogForwardingProfile",
+        "objects.DynamicUserGroup",
+        "objects.Region",
+        "objects.Edl",
         "firewall.Firewall",
         "panorama.DeviceGroup",
         "panorama.Template",
         "panorama.TemplateStack",
+        "plugins.CloudServicesPlugin",
+        "policies.Rulebase",
     )
+    OPSTATES = {
+        "dg_hierarchy": PanoramaDeviceGroupHierarchy,
+    }
 
     def __init__(
         self,
@@ -422,9 +474,13 @@ class Panorama(base.PanDevice):
         cmd_xml=True,
         extra_qs=None,
         retry_on_peer=False,
+        quote='"',
     ):
         """Perform operational command on this Panorama
 
+        Operational commands are most any command that is not a debug or config
+        command. These include many 'show' commands such as ``show system info``.
+
         Args:
             cmd (str): The operational command to execute
             vsys (str): Ignored for Panorama
@@ -432,6 +488,7 @@ class Panorama(base.PanDevice):
             cmd_xml (bool): True: cmd is not XML, False: cmd is XML (Default: True)
             extra_qs: Extra parameters for API call
             retry_on_peer (bool): Try on active Firewall first, then try on passive Firewall
+            quote (str): The quote character when the supplied `cmd` is a string and `cmd_xml=True`
 
         Returns:
             xml.etree.ElementTree: The result of the operational command. May also return a string of XML if xml=True
@@ -445,6 +502,7 @@ class Panorama(base.PanDevice):
             cmd_xml=cmd_xml,
             extra_qs=extra_qs,
             retry_on_peer=retry_on_peer,
+            quote=quote,
         )
 
     def xpath_vsys(self):
@@ -825,46 +883,25 @@ class Panorama(base.PanDevice):
 
         return ans
 
-    def _setup_opstate(self):
-        self.opstate = PanoramaOpState(self)
+    def refresh_shared_objects(self, running_config=False, exceptions=True):
+        """Refresh all objects present in the shared scope.
 
+        NOTE:  This will remove any instance of any class that shows up
+        in the :class:`panos.panorama.Panorama` `CHILDTYPES` constant from
+        this Panorama's `children`.
 
-class PanoramaOpState(object):
-    """Panorama OP state handling."""
-
-    def __init__(self, obj):
-        self.dg_hierarchy = PanoramaDeviceGroupHierarchy(obj)
-
-
-class PanoramaDeviceGroupHierarchy(object):
-    """Operational state handling for device group hierarchy."""
-
-    def __init__(self, obj):
-        self.obj = obj
-        self.parent = None
-
-    def fetch(self):
-        """Returns a dict of device groups and their parents.
-
-        Keys in the dict are the device group's name, while the value is the
-        name of that device group's parent.  Top level device groups will have
-        a parent of ``None``.
-
-        Returns:
-            dict
+        Args:
+            running_config (bool): Set to True to refresh from the running
+                configuration (Default: False)
+            exceptions (bool): Set to False to prevent exceptions on failure
+                (Default: True)
 
         """
+        xml = self.xapi.get("/config/shared")
+        if xml is not None:
+            xml = xml.find("./result/shared")
 
-        resp = self.obj.op("show dg-hierarchy")
-        data = resp.find("./result/dg-hierarchy")
-
-        ans = {}
-        nodes = [(None, x) for x in data.findall("./dg")]
-        for parent, elm in iter(nodes):
-            ans[elm.attrib["name"]] = parent
-            nodes.extend((elm.attrib["name"], x) for x in elm.findall("./dg"))
-
-        return ans
+        self.refresh(xml=xml, running_config=running_config, exceptions=exceptions)
 
 
 class PanoramaCommit(object):
