@@ -18,17 +18,151 @@
 """Policies module contains policies and rules that exist in the 'Policies' tab in the firewall GUI"""
 
 import xml.etree.ElementTree as ET
-from datetime import datetime
 
 import panos.errors as err
 from panos import getlogger
-from panos.base import ENTRY, MEMBER, PanObject, Root
+from panos.base import ENTRY, MEMBER, OpState, PanObject, Root
 from panos.base import VarPath as Var
 from panos.base import VersionedPanObject, VersionedParamPath
 
 logger = getlogger(__name__)
 
 
+class HitCount(OpState):
+    """Hit count operational data."""
+
+    FIELDS = (
+        ("latest", "latest", "str"),
+        ("hit_count", "hit-count", "int"),
+        ("last_hit_timestamp", "last-hit-timestamp", "int"),
+        ("last_reset_timestamp", "last-reset-timestamp", "int"),
+        ("first_hit_timestamp", "first-hit-timestamp", "int"),
+        ("rule_creation_timestamp", "rule-creation-timestamp", "int"),
+        ("rule_modification_timestamp", "rule-modification-timestamp", "int"),
+    )
+
+    def _setup(self, name=None, elm=None):
+        self.name = name if self.obj is None else self.obj.uid
+        self._refresh_xml(elm)
+
+    def refresh(self, elm=None):
+        if elm is None and self.obj is not None:
+            self.obj.parent.opstate.hit_count.refresh(
+                self.obj.HIT_COUNT_STYLE, [self.name,],
+            )
+        else:
+            self._refresh_xml(elm)
+
+    def _refresh_xml(self, elm):
+        for param, path, param_type in self.FIELDS:
+            if param_type == "int":
+                setattr(self, param, self._int(elm, path))
+            else:
+                setattr(self, param, self._str(elm, path))
+
+
+class AuditCommentLog(OpState):
+    """A single audit comment log entry."""
+
+    def __init__(self, elm):
+        self.admin = self._str(elm, "admin")
+        self.comment = self._str(elm, "comment")
+        self.config_version = self._int(elm, "config_ver")
+        self.time = self._datetime(elm, "time_generated", "%Y/%m/%d %H:%M:%S")
+
+
+class RulebaseHitCount(OpState):
+    """Operational state handling for rulebase hit counts."""
+
+    def refresh(self, style, rules=None, all_rules=False):
+        """Retrieves hit count information for the specified rules.
+
+        PAN-OS 8.1+
+
+        Args:
+            style (str): The rule style to use.  The style can be
+                "application-override", "authentication", "decryption", "dos",
+                "nat", "pbf", "qos", "sdwan", "security", or "tunnel-inspect".
+            rules (list): A list of rules.  This can be a mix of `panos.policies`
+                instances or basic strings.  If no rules are given, then the hit
+                count for all rules is retrieved.
+            all_rules (bool): If this is False, only retrieve hit count information
+                for the rules attached to the rulebase of the specified style.  If
+                this is True, then get all rules.  Either way, any rule whose hit count
+                is retrieved and is in the object hierarchy has the hit count data
+                saved to its `opstate`.
+
+        Returns:
+            dict:  A dict where the key is the rule name and the value is the hit count information.
+
+        """
+        dev = self.obj.nearest_pandevice()
+
+        if dev.retrieve_panos_version() < (8, 1, 0):
+            raise err.PanDeviceError("Rule hit count is supported in PAN-OS 8.1+")
+
+        kids = []
+        for x in self.obj.children:
+            if not hasattr(x, "HIT_COUNT_STYLE") or x.HIT_COUNT_STYLE != style:
+                continue
+            if rules is None or x.uid in rules:
+                kids.append(x)
+
+        cmd = ET.Element("show")
+        sub = ET.SubElement(cmd, "rule-hit-count")
+        res_path = "./result/rule-hit-count"
+        rb_type_path_map = {PreRulebase: "pre-rulebase", PostRulebase: "post-rulebase"}
+
+        if dev.__class__.__name__ == "Panorama":
+            if self.obj.parent.__class__.__name__ == "Panorama":
+                sub = ET.SubElement(sub, "shared")
+                res_path += "/shared"
+            elif self.obj.parent.__class__.__name__ == "DeviceGroup":
+                sub = ET.SubElement(sub, "device-group")
+                sub = ET.SubElement(sub, "entry", {"name": self.obj.parent.name})
+                res_path += "/device-group/entry"
+            sub = ET.SubElement(sub, rb_type_path_map.get(type(self.obj)))
+        else:
+            sub = ET.SubElement(sub, "vsys")
+            sub = ET.SubElement(sub, "vsys-name")
+            sub = ET.SubElement(sub, "entry", {"name": dev.vsys or "vsys1"})
+            sub = ET.SubElement(sub, "rule-base")
+            res_path += "/vsys/entry"
+
+        sub = ET.SubElement(sub, "entry", {"name": style})
+        sub = ET.SubElement(sub, "rules")
+        res_path += "/rule-base/entry/rules/entry"
+
+        if all_rules:
+            ET.SubElement(sub, "all")
+        else:
+            # Loop over rules specified or the object hierarchy.
+            rule_list = rules or kids or []
+            if not rule_list:
+                return {}
+            sub = ET.SubElement(sub, "list")
+            for x in rule_list:
+                if hasattr(x, "uid"):
+                    ET.SubElement(sub, "member").text = x.uid
+                else:
+                    ET.SubElement(sub, "member").text = x
+
+        res = dev.op(ET.tostring(cmd, encoding="utf-8"), cmd_xml=False)
+        ans = {}
+
+        for elm in res.findall(res_path):
+            name = elm.attrib["name"]
+            for x in kids:
+                if x.uid == name:
+                    x.opstate.hit_count.refresh(elm)
+                    ans[name] = x.opstate.hit_count
+                    break
+            else:
+                ans[name] = HitCount(None, name=name, elm=elm)
+
+        return ans
+
+
 class Rulebase(VersionedPanObject):
     """Rulebase for a Firewall
 
@@ -39,19 +173,21 @@ class Rulebase(VersionedPanObject):
 
     NAME = None
     ROOT = Root.VSYS
+    OPSTATES = {
+        "hit_count": RulebaseHitCount,
+    }
     CHILDTYPES = (
         "policies.NatRule",
         "policies.PolicyBasedForwarding",
         "policies.SecurityRule",
         "policies.DecryptionRule",
+        "policies.ApplicationOverride",
+        "policies.AuthenticationRule",
     )
 
     def _setup(self):
         self._xpaths.add_profile(value="/rulebase")
 
-    def _setup_opstate(self):
-        self.opstate = RulebaseOpState(self)
-
 
 class PreRulebase(Rulebase):
     """Pre-rulebase for a Panorama
@@ -75,6 +211,97 @@ class PostRulebase(Rulebase):
         self._xpaths.add_profile(value="/post-rulebase")
 
 
+class RuleAuditComment(OpState):
+    """Operational state handling for a rule's audit comments.
+
+    Note:  Audit comments are present in PAN-OS 9.0+.
+
+    """
+
+    def history(self, count=100, direction="backward", skip=None):
+        """Returns a chunk of historical audit comment logs.
+
+        Args:
+            count (int): Number of audit comments to return, maximum 5000.
+            direction (str): Specify whether logs are shown oldest first
+                (``forward``) or newest first (``backward``).
+            skip (int): Specify the number of logs to skip when doing log
+                retrieval.  This is useful when retrieving logs in batches
+                where you can skip the previously retrieved logs.
+
+        Returns:
+            list of :class:`panos.policies.AuditCommentLog`
+
+        """
+        dev = self.obj.nearest_pandevice()
+
+        # Build up the query.
+        query = "(subtype eq audit-comment)"
+        # Add in the name.
+        query += " and (path contains '\\'{0}\\'')".format(self.obj.uid)
+        # Add in the rule type.
+        query += " and (path contains '{0}')".format(
+            self.obj.xpath_short().split("/")[-2],
+        )
+        # Add in the rulebase type.
+        p = self.obj.parent
+        if p is None:
+            raise err.PanDeviceError("rule has empty parent")
+        if not any(isinstance(p, x) for x in (Rulebase, PreRulebase, PostRulebase)):
+            raise err.PanDeviceError("{0} has non-rulebase parent".format(self.obj.uid))
+        query += " and (path contains {0})".format(p.xpath().split("/")[-1])
+        # Add in the vsys or device group.
+        query += " and (path contains '\\'{0}\\'')".format(p.vsys)
+
+        extra_qs = {
+            "dir": direction,
+            "uniq": "yes",
+        }
+        if skip is not None:
+            extra_qs["skip"] = "{0}".format(int(skip))
+
+        resp = dev.xapi.log("config", count, filter=query, extra_qs=extra_qs)
+
+        return [AuditCommentLog(x) for x in resp.findall("./result/log/logs/entry")]
+
+    def current(self):
+        """Returns the current audit comment.
+
+        Returns:
+            string
+
+        """
+        cmd = ET.Element("show")
+        sub = ET.SubElement(cmd, "config")
+        sub = ET.SubElement(sub, "list")
+        sub = ET.SubElement(sub, "audit-comments")
+        ET.SubElement(sub, "xpath").text = self.obj.xpath()
+
+        ans = self.obj.nearest_pandevice().op(
+            ET.tostring(cmd, encoding="utf-8"), cmd_xml=False,
+        )
+
+        resp = ans.find("./result/entry/comment")
+        if resp is not None:
+            return resp.text
+
+    def update(self, comment):
+        """Sets an audit comment for the given rule.
+
+        Args:
+            comment (str): The audit comment.
+
+        """
+        cmd = ET.Element("set")
+        sub = ET.SubElement(cmd, "audit-comment")
+        ET.SubElement(sub, "xpath").text = self.obj.xpath()
+        ET.SubElement(sub, "comment").text = comment
+
+        self.obj.nearest_pandevice().op(
+            ET.tostring(cmd, encoding="utf-8"), cmd_xml=False,
+        )
+
+
 class SecurityRule(VersionedPanObject):
     """Security Rule
 
@@ -84,7 +311,7 @@ class SecurityRule(VersionedPanObject):
         tozone (list): To zones
         source (list): Source addresses
         source_user (list): Source users and groups
-        hip_profiles (list): GlobalProtect host integrity profiles
+        hip_profiles (list): (PAN-OS 10.1.4-) GlobalProtect host integrity profiles
         destination (list): Destination addresses
         application (list): Applications
         service (list): Destination services (ports) (Default:
@@ -133,6 +360,10 @@ class SecurityRule(VersionedPanObject):
     SUFFIX = ENTRY
     ROOT = Root.VSYS
     HIT_COUNT_STYLE = "security"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
 
     def _setup(self):
         # xpaths
@@ -157,6 +388,14 @@ class SecurityRule(VersionedPanObject):
                 )
             )
 
+        # 10.1.5 drops support for hip-profiles,
+        # so we want to make sure we don't include it in the request
+        # body that we send to the api
+        for param in params:
+            if param.name == "hip_profiles":
+                param.add_profile("10.1.5", exclude=True)
+                break
+
         params.append(
             VersionedParamPath(
                 "service",
@@ -242,9 +481,6 @@ class SecurityRule(VersionedPanObject):
 
         self._params = tuple(params)
 
-    def _setup_opstate(self):
-        self.opstate = RuleOpState(self)
-
 
 class NatRule(VersionedPanObject):
     """NAT Rule
@@ -318,6 +554,10 @@ class NatRule(VersionedPanObject):
     SUFFIX = ENTRY
     ROOT = Root.VSYS
     HIT_COUNT_STYLE = "nat"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
 
     def _setup(self):
         # xpaths
@@ -603,8 +843,84 @@ class NatRule(VersionedPanObject):
 
         self._params = tuple(params)
 
-    def _setup_opstate(self):
-        self.opstate = RuleOpState(self)
+
+class ApplicationOverride(VersionedPanObject):
+    """ApplicationOverride
+
+    Args:
+        name (str): Name of the rule
+        fromzone (list): From zones
+        tozone (list): To zones
+        source (list): Source addresses
+        destination (list): Destination addresses
+        application (str): Applications
+        description (str): Description of this rule
+        tag (list): Administrative tags
+        negate_source (bool): Match on the reverse of the 'source' attribute
+        negate_destination (bool): Match on the reverse of the 'destination'
+            attribute
+        disabled (bool): Disable this rule
+        negate_target (bool): Target all but the listed target firewalls
+            (applies to panorama/device groups only)
+        target (list): Apply this policy to the listed firewalls only
+            (applies to panorama/device groups only)
+        port (str): Destination port
+        protocol (str): Protocol used
+        group_tag (str): (PAN-OS 9.0+) The group tag.
+
+    """
+
+    SUFFIX = ENTRY
+    ROOT = Root.VSYS
+    HIT_COUNT_STYLE = "application-override"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
+
+    def _setup(self):
+        # xpaths
+        self._xpaths.add_profile(value="/application-override/rules")
+
+        # params
+        params = []
+
+        any_defaults = (
+            ("fromzone", "from"),
+            ("tozone", "to"),
+            ("source", "source"),
+            ("destination", "destination"),
+        )
+        for var_name, path in any_defaults:
+            params.append(
+                VersionedParamPath(
+                    var_name, default=["any",], vartype="member", path=path
+                )
+            )
+        params.append(VersionedParamPath("application", path="application"))
+        params.append(VersionedParamPath("description", path="description"))
+        params.append(VersionedParamPath("tag", path="tag", vartype="member"))
+        params.append(
+            VersionedParamPath("negate_source", path="negate-source", vartype="yesno")
+        )
+        params.append(
+            VersionedParamPath(
+                "negate_destination", path="negate-destination", vartype="yesno"
+            )
+        )
+        params.append(VersionedParamPath("disabled", path="disabled", vartype="yesno"))
+        params.append(
+            VersionedParamPath("negate_target", path="target/negate", vartype="yesno")
+        )
+        params.append(
+            VersionedParamPath("target", path="target/devices", vartype="entry")
+        )
+        params.append(VersionedParamPath("port", path="port"))
+        params.append(VersionedParamPath("protocol", path="protocol"))
+        params.append(VersionedParamPath("group_tag", exclude=True))
+        params[-1].add_profile("9.0.0", path="group-tag")
+
+        self._params = tuple(params)
 
 
 class PolicyBasedForwarding(VersionedPanObject):
@@ -656,6 +972,10 @@ class PolicyBasedForwarding(VersionedPanObject):
     SUFFIX = ENTRY
     ROOT = Root.VSYS
     HIT_COUNT_STYLE = "pbf"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
 
     def _setup(self):
         # xpaths
@@ -796,9 +1116,6 @@ class PolicyBasedForwarding(VersionedPanObject):
 
         self._params = tuple(params)
 
-    def _setup_opstate(self):
-        self.opstate = RuleOpState(self)
-
 
 class DecryptionRule(VersionedPanObject):
     """Decryption rule.
@@ -834,16 +1151,24 @@ class DecryptionRule(VersionedPanObject):
             handshakes.
         log_failed_tls_handshakes (bool): (PAN-OS 10.0+) Log failed TLS handshakes.
         log_setting (str): (PAN-OS 10.0+) Log setting.
+        negate_target (bool): Target all but the listed target firewalls
+            (applies to panorama/device groups only)
+        target (list): Apply this policy to the listed firewalls only
+            (applies to panorama/device groups only)
 
     """
 
     SUFFIX = ENTRY
     ROOT = Root.VSYS
     HIT_COUNT_STYLE = "decryption"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
 
     def _setup(self):
         # xpaths
-        self._xpaths.add_profile(value="/pbf/rules")
+        self._xpaths.add_profile(value="/decryption/rules")
 
         # params
         params = []
@@ -940,267 +1265,131 @@ class DecryptionRule(VersionedPanObject):
         params[-1].add_profile(
             "10.0.0", path="log-setting",
         )
+        params.append(
+            VersionedParamPath("negate_target", path="target/negate", vartype="yesno")
+        )
+        params.append(
+            VersionedParamPath("target", path="target/devices", vartype="entry")
+        )
 
         self._params = tuple(params)
 
-    def _setup_opstate(self):
-        self.opstate = RuleOpState(self)
 
+class AuthenticationRule(VersionedPanObject):
+    """Authentication Rule
 
-class RulebaseOpState(object):
-    """Operational state handling for rulebase classes."""
+    Both the naming convention and the order of the parameters tries to closly
+    match what is presented in the GUI.
 
-    def __init__(self, obj):
-        self.hit_count = RulebaseHitCount(obj)
-
-
-class RulebaseHitCount(object):
-    """Operational state handling for rulebase hit counts."""
-
-    def __init__(self, obj):
-        self.obj = obj
-
-    def refresh(self, style, rules=None, all_rules=False):
-        """Retrieves hit count information for the specified rules.
-
-        PAN-OS 8.1+
-
-        Args:
-            style (str): The rule style to use.  The style can be
-                "application-override", "authentication", "decryption", "dos",
-                "nat", "pbf", "qos", "sdwan", "security", or "tunnel-inspect".
-            rules (list): A list of rules.  This can be a mix of `panos.policies`
-                instances or basic strings.  If no rules are given, then the hit
-                count for all rules is retrieved.
-            all_rules (bool): If this is False, only retrieve hit count information
-                for the rules attached to the rulebase of the specified style.  If
-                this is True, then get all rules.  Either way, any rule whose hit count
-                is retrieved and is in the object hierarchy has the hit count data
-                saved to its `opstate`.
-
-        Returns:
-            dict:  A dict where the key is the rule name and the value is the hit count information.
-
-        """
-        dev = self.obj.nearest_pandevice()
-
-        if dev.retrieve_panos_version() < (8, 1, 0):
-            raise err.PanDeviceError("Rule hit count is supported in PAN-OS 8.1+")
-
-        kids = []
-        for x in self.obj.children:
-            if not hasattr(x, "HIT_COUNT_STYLE") or x.HIT_COUNT_STYLE != style:
-                continue
-            if rules is None or x.uid in rules:
-                kids.append(x)
-
-        cmd = ET.Element("show")
-        sub = ET.SubElement(cmd, "rule-hit-count")
-        sub = ET.SubElement(sub, "vsys")
-        sub = ET.SubElement(sub, "vsys-name")
-        sub = ET.SubElement(sub, "entry", {"name": dev.vsys or "vsys1"})
-        sub = ET.SubElement(sub, "rule-base")
-        sub = ET.SubElement(sub, "entry", {"name": style})
-        sub = ET.SubElement(sub, "rules")
-
-        if all_rules:
-            ET.SubElement(sub, "all")
-        else:
-            # Loop over rules specified or the object hierarchy.
-            rule_list = rules or kids or []
-            if not rule_list:
-                return {}
-            sub = ET.SubElement(sub, "list")
-            for x in rule_list:
-                if hasattr(x, "uid"):
-                    ET.SubElement(sub, "member").text = x.uid
-                else:
-                    ET.SubElement(sub, "member").text = x
-
-        res = dev.op(ET.tostring(cmd, encoding="utf-8"), cmd_xml=False)
-
-        ans = {}
-        for elm in res.findall(
-            "./result/rule-hit-count/vsys/entry/rule-base/entry/rules/entry"
-        ):
-            name = elm.attrib["name"]
-            for x in kids:
-                if x.uid == name:
-                    x.opstate.hit_count.refresh(elm)
-                    ans[name] = x.opstate.hit_count
-                    break
-            else:
-                ans[name] = HitCount(name=name, elm=elm)
-
-        return ans
-
-
-class OpStateObject(object):
-    """A container object for opstate data."""
-
-    def _str(self, elm, field):
-        if elm is not None:
-            val = elm.find("./{0}".format(field))
-            if val is not None:
-                return val.text
-
-    def _int(self, elm, field):
-        val = self._str(elm, field)
-        if val is not None:
-            return int(val)
-
-    def _datetime(self, elm, field, fmt):
-        val = self._str(elm, field)
-        if val is not None:
-            try:
-                return datetime.strptime(val, fmt)
-            except ValueError:
-                pass
-            return val
-
-
-class HitCount(OpStateObject):
-    """Hit count operational data."""
-
-    FIELDS = (
-        ("latest", "latest", "str"),
-        ("hit_count", "hit-count", "int"),
-        ("last_hit_timestamp", "last-hit-timestamp", "int"),
-        ("last_reset_timestamp", "last-reset-timestamp", "int"),
-        ("first_hit_timestamp", "first-hit-timestamp", "int"),
-        ("rule_creation_timestamp", "rule-creation-timestamp", "int"),
-        ("rule_modification_timestamp", "rule-modification-timestamp", "int"),
-    )
-
-    def __init__(self, obj=None, name=None, elm=None):
-        self.obj = obj
-        self.name = name if obj is None else obj.uid
-        self._refresh_xml(elm)
-
-    def refresh(self, elm=None):
-        if elm is None and self.obj is not None:
-            self.obj.parent.opstate.hit_count.refresh(
-                self.obj.HIT_COUNT_STYLE, [self.name,],
-            )
-        else:
-            self._refresh_xml(elm)
-
-    def _refresh_xml(self, elm):
-        for param, path, param_type in self.FIELDS:
-            if param_type == "int":
-                setattr(self, param, self._int(elm, path))
-            else:
-                setattr(self, param, self._str(elm, path))
-
-
-class RuleOpState(object):
-    """Operational state handling for a rule in the rulebase."""
-
-    def __init__(self, obj):
-        self.obj = obj
-        self.audit_comment = RuleAuditComment(obj)
-        self.hit_count = HitCount(obj)
-
-
-class RuleAuditComment(object):
-    """Operational state handling for a rule's audit comments.
-
-    Note:  Audit comments are present in PAN-OS 9.0+.
+    Args:
+        name (str): The name
+        description (str): The description
+        uuid (str): (PAN-OS 9.0+) The UUID for this rule.
+        source_zones (list): The source zones.
+        source_addresses (list): The source addresses.
+        negate_source (bool): Negate the source addresses.
+        destination_zones (list): The destination zones.
+        destination_addresses (list): The destination addresses.
+        negate_destination (bool): Negate the destination addresses.
+        tag (list): Administrative tags
+        disabled (bool): Disable this rule
+        service (str): The service
+        source_hip (list): (PAN-OS 10.0+) The source HIP info.
+        source_users (list): The source users.
+        url_categories (list): URL categories.
+        group_tag (str): (PAN-OS 9.0+) The group tag.
+        authentication_enforcement (str): The authentication enforcement object.
+        timeout (str): The authentication timeout.
+        negate_target (bool): Target all but the listed target firewalls,
+            (applies to panorama/device groups only)
+        target (list): Apply this policy to the listed firewalls only,
+            (applies to panorama/device groups only)
+        log_setting (str): (PAN-OS 10.0+) Log setting.
+        log_authentication_timeout (bool): Whether the rules logs authentication timeouts or not.
 
     """
 
-    def __init__(self, obj):
-        self.obj = obj
+    SUFFIX = ENTRY
+    ROOT = Root.VSYS
+    HIT_COUNT_STYLE = "authentication"
+    OPSTATES = {
+        "audit_comment": RuleAuditComment,
+        "hit_count": HitCount,
+    }
 
-    def history(self, count=100, direction="backward", skip=None):
-        """Returns a chunk of historical audit comment logs.
+    def _setup(self):
+        # xpaths
+        self._xpaths.add_profile(value="/authentication/rules")
 
-        Args:
-            count (int): Number of audit comments to return, maximum 5000.
-            direction (str): Specify whether logs are shown oldest first
-                (``forward``) or newest first (``backward``).
-            skip (int): Specify the number of logs to skip when doing log
-                retrieval.  This is useful when retrieving logs in batches
-                where you can skip the previously retrieved logs.
+        # params
+        params = []
 
-        Returns:
-            list of :class:`panos.policies.AuditCommentLog`
-
-        """
-        dev = self.obj.nearest_pandevice()
-
-        # Build up the query.
-        query = "(subtype eq audit-comment)"
-        # Add in the name.
-        query += " and (path contains '\\'{0}\\'')".format(self.obj.uid)
-        # Add in the rule type.
-        query += " and (path contains '{0}')".format(
-            self.obj.xpath_short().split("/")[-2],
+        params.append(VersionedParamPath("description", path="description"))
+        params.append(VersionedParamPath("uuid", exclude=True))
+        params[-1].add_profile("9.0.0", vartype="attrib", path="uuid")
+        params.append(
+            VersionedParamPath("source_zones", vartype="member", path="from",)
         )
-        # Add in the rulebase type.
-        p = self.obj.parent
-        if p is None:
-            raise err.PanDeviceError("rule has empty parent")
-        if not any(isinstance(p, x) for x in (Rulebase, PreRulebase, PostRulebase)):
-            raise err.PanDeviceError("{0} has non-rulebase parent".format(self.obj.uid))
-        query += " and (path contains {0})".format(p.xpath().split("/")[-1])
-        # Add in the vsys or device group.
-        query += " and (path contains '\\'{0}\\'')".format(p.vsys)
-
-        extra_qs = {
-            "dir": direction,
-            "uniq": "yes",
-        }
-        if skip is not None:
-            extra_qs["skip"] = "{0}".format(int(skip))
-
-        resp = dev.xapi.log("config", count, filter=query, extra_qs=extra_qs)
-
-        return [AuditCommentLog(x) for x in resp.findall("./result/log/logs/entry")]
-
-    def current(self):
-        """Returns the current audit comment.
-
-        Returns:
-            string
-
-        """
-        cmd = ET.Element("show")
-        sub = ET.SubElement(cmd, "config")
-        sub = ET.SubElement(sub, "list")
-        sub = ET.SubElement(sub, "audit-comments")
-        ET.SubElement(sub, "xpath").text = self.obj.xpath()
-
-        ans = self.obj.nearest_pandevice().op(
-            ET.tostring(cmd, encoding="utf-8"), cmd_xml=False,
+        params.append(
+            VersionedParamPath(
+                "source_addresses", default=["any",], vartype="member", path="source"
+            )
+        )
+        params.append(
+            VersionedParamPath("negate_source", vartype="yesno", path="negate-source",)
+        )
+        params.append(
+            VersionedParamPath("destination_zones", vartype="member", path="to",)
+        )
+        params.append(
+            VersionedParamPath(
+                "destination_addresses",
+                default=["any",],
+                vartype="member",
+                path="destination",
+            )
+        )
+        params.append(
+            VersionedParamPath(
+                "negate_destination", vartype="yesno", path="negate-destination",
+            )
+        )
+        params.append(VersionedParamPath("tag", vartype="member", path="tag",))
+        params.append(VersionedParamPath("disabled", vartype="yesno", path="disabled",))
+        params.append(VersionedParamPath("service", vartype="member", path="service",))
+        params.append(VersionedParamPath("source_hip", exclude=True,))
+        params[-1].add_profile(
+            "10.0.0", path="source-hip", vartype="member",
+        )
+        params.append(
+            VersionedParamPath("source_users", vartype="member", path="source-user")
+        )
+        params.append(
+            VersionedParamPath("url_categories", vartype="member", path="category",)
+        )
+        params.append(VersionedParamPath("group_tag", exclude=True,))
+        params[-1].add_profile(
+            "9.0.0", path="group-tag",
+        )
+        params.append(
+            VersionedParamPath(
+                "authentication_enforcement", path="authentication-enforcement",
+            )
+        )
+        params.append(VersionedParamPath("timeout", path="timeout",))
+        params.append(
+            VersionedParamPath("negate_target", path="target/negate", vartype="yesno")
+        )
+        params.append(
+            VersionedParamPath("target", path="target/devices", vartype="entry")
+        )
+        params.append(VersionedParamPath("log_setting", path="log-setting"))
+        params.append(
+            VersionedParamPath(
+                "log_authentication_timeout",
+                path="log-authentication-timeout",
+                vartype="yesno",
+            )
         )
 
-        resp = ans.find("./result/entry/comment")
-        if resp is not None:
-            return resp.text
-
-    def update(self, comment):
-        """Sets an audit comment for the given rule.
-
-        Args:
-            comment (str): The audit comment.
-
-        """
-        cmd = ET.Element("set")
-        sub = ET.SubElement(cmd, "audit-comment")
-        ET.SubElement(sub, "xpath").text = self.obj.xpath()
-        ET.SubElement(sub, "comment").text = comment
-
-        self.obj.nearest_pandevice().op(
-            ET.tostring(cmd, encoding="utf-8"), cmd_xml=False,
-        )
-
-
-class AuditCommentLog(OpStateObject):
-    """A single audit comment log entry."""
-
-    def __init__(self, elm):
-        self.admin = self._str(elm, "admin")
-        self.comment = self._str(elm, "comment")
-        self.config_version = self._int(elm, "config_ver")
-        self.time = self._datetime(elm, "time_generated", "%Y/%m/%d %H:%M:%S")
+        self._params = tuple(params)
